from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from http.client import HTTPException
from os import cpu_count

from mureq import Response, TooManyRedirects, get
from trafilatura import extract

from safari_to_sqlite.constants import ScrapeStatus
from safari_to_sqlite.errors import FailedDownloadError

BATCH_SIZE = (cpu_count() if cpu_count() is not None else 6) * 2


def _download(url: str) -> str:
    response: Response = get(url, max_redirects=2)
    if not response.ok:
        raise FailedDownloadError(response.status_code)
    try:
        return response.content.decode()
    except UnicodeDecodeError as e:
        raise FailedDownloadError(ScrapeStatus.UnicodeFailed.value) from e


def _extract_body(url_tuple: tuple[str]) -> tuple[str, str | int]:
    url = url_tuple[0]
    try:
        body = _download(url)
        extracted: str = extract(
            body,
            output_format="markdown",
            favor_recall=True,
            include_links=True,
        )
    except FailedDownloadError as e:
        return url, e.code
    except TooManyRedirects:
        return url, ScrapeStatus.RedirectsExceeded.value
    except TimeoutError:
        return url, ScrapeStatus.Timeout.value
    except HTTPException:
        return url, ScrapeStatus.UnknownHttp.value
    else:
        return url, extracted


def extract_bodies(
    urls: Iterable[tuple[str]],
) -> tuple[list[tuple[str, str]], list[tuple[str, int]]]:
    """Download and extract bodies from URLs."""
    successes: list[tuple[str, str]] = []
    errors: list[tuple[str, int]] = []
    with ThreadPoolExecutor(max_workers=BATCH_SIZE) as executor:
        # Mapping the read_first_line function over all file paths
        for result in executor.map(_extract_body, urls):
            if isinstance(result[1], int):
                errors.append(result)  # pytype: disable=container-type-mismatch
            else:
                successes.append(result)
    return successes, errors



from enum import Enum

BODY = "body"
BROWSER = "browser"
FIRST_SEEN = "firstSeen"
HOST = "host"  # refers to user's local hostname NOT the URL host
SCRAPE_STATUS = "scrapeStatus"
TABS = "tabs"
TAB_INDEX = "tabIndex"
TITLE = "title"
URL = "url"
WINDOW_ID = "windowId"

TURSO_SAFARI = "turso_safari"
TURSO_URL = "turso_url"
TURSO_AUTH_TOKEN = "turso_auth_token"  # noqa: S105


class ScrapeStatus(int, Enum):
    """Enum for scrape status."""

    NotScraped = -1
    ExtractFailed = -2
    UnicodeFailed = -3
    RedirectsExceeded = -4
    Timeout = -5
    UnknownHttp = -6


class Browser(str, Enum):
    """Enum for browser."""

    Safari = "safari"
    Icloud = "icloud"
    Chrome = "chrome"


SEP = "150r2M72e7D6Lb7Z9u1g4BSQBonv6U21W1fmX8B1TXR8XXqB2wTgQzqwoB06144d"



"""Stripped down version of the more_itertools module.

https://github.com/more-itertools/more-itertools
"""

from functools import partial
from itertools import islice


def take(n, iterable):
    """Return first *n* items of the iterable as a list.

        >>> take(3, range(10))
        [0, 1, 2]

    If there are fewer than *n* items in the iterable, all of them are
    returned.

        >>> take(10, range(3))
        [0, 1, 2]

    """
    return list(islice(iterable, n))


def chunked(iterable, n, strict=False):
    """Break *iterable* into lists of length *n*:

        >>> list(chunked([1, 2, 3, 4, 5, 6], 3))
        [[1, 2, 3], [4, 5, 6]]

    By the default, the last yielded list will have fewer than *n* elements
    if the length of *iterable* is not divisible by *n*:

        >>> list(chunked([1, 2, 3, 4, 5, 6, 7, 8], 3))
        [[1, 2, 3], [4, 5, 6], [7, 8]]

    To use a fill-in value instead, see the :func:`grouper` recipe.

    If the length of *iterable* is not divisible by *n* and *strict* is
    ``True``, then ``ValueError`` will be raised before the last
    list is yielded.

    """
    iterator = iter(partial(take, n, iter(iterable)), [])
    if strict:
        if n is None:
            raise ValueError("n must not be None when using strict mode.")

        def ret():
            for chunk in iterator:
                if len(chunk) != n:
                    raise ValueError("iterable is not divisible by n.")
                yield chunk

        return iter(ret())
    else:
        return iterator



import sqlite3
from collections.abc import Iterable
from pathlib import Path

from safari_to_sqlite.constants import Browser
from safari_to_sqlite.datastore import TabRow


def read_history() -> Iterable[TabRow]:
    """Read Safari history from local SQLite database."""
    db_path = Path.home() / "Library/Safari/History.db"
    con = sqlite3.connect(db_path)
    query = (
        "SELECT history_items.url, title, visit_time, history_visits.origin "
        "FROM history_visits LEFT JOIN history_items "
        "ON history_visits.history_item=history_items.id;"
    )
    cur = con.cursor()
    for row in cur.execute(query):
        yield TabRow(
            url=row[0],
            title=row[1] or "",
            body="",
            window_id=-1,
            tab_index=-1,
            host="",
            first_seen=int(row[2]),
            scrape_status=-1,
            browser=Browser.Safari.value if row[3] == 1 else Browser.Icloud.value,
        )
    cur.close()
    con.close()
